#!/usr/bin/env python3

#This script allows the user to pull the existing gottcha database
import requests
import sys
import os
import tarfile
from tqdm import *
import argparse
import pkg_resources
import json
from gottcha import GOTTCHA_DB_LATEST, FILE_NAME



def parse_params(args):
    parser = argparse.ArgumentParser(prog='pull_database.py', description="""This script will pull the latest version of the Gottcha2 database.""")

    parser.add_argument('-u', '--url', default=GOTTCHA_DB_LATEST,
                    help='specify a URL to pull from (will override the default)')
    parser.add_argument('-r', '--rank', default='species',
                    help='taxonomic rank of the database (superkingdom, phylum, class, order, famiily, genus, species)')
    return parser.parse_args(args)

def download_db(argvs):
    if os.path.isdir('database'):
        sys.exit('Please make sure a database directory does not exist.')

    os.mkdir('database')

    if argvs.url:
        GOTTCHA_DB_LATEST = argvs.url
    with requests.get(GOTTCHA_DB_LATEST.replace('species',argvs.rank,1), stream=True) as r:
        r.raise_for_status()
        with open(FILE_NAME, 'wb') as f:
            pbar = tqdm(total=int(r.headers['Content-Length']),unit='B', unit_scale=True, unit_divisor=1024)
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
                    pbar.update(len(chunk))
    tar = tarfile.open(FILE_NAME.replace('species',argvs.rank,1))
    tar.extractall('database')
    tar.close()


def main(args):
    argvs = parse_params(args)
    download_db(argvs)

if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3

__author__    = "Po-E (Paul) Li, Bioscience Division, Los Alamos National Laboratory"
__credits__   = ["Po-E Li", "Anna Chernikov", "Jason Gans", "Tracey Freites", "Patrick Chain"]
__version__   = "2.1.8.6"
__date__      = "2018/10/07"
__copyright__ = """
Copyright (2019). Traid National Security, LLC. This material was produced
under U.S. Government contract DE-AC52-06NA25396 for Los Alamos National Laboratory
(LANL), which is operated by Los Alamos National Security, LLC for the U.S.
Department of Energy. The U.S. Government has rights to use, reproduce, and
distribute this software.  NEITHER THE GOVERNMENT NOR LOS ALAMOS NATIONAL SECURITY,
LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE
USE OF THIS SOFTWARE.  If software is modified to produce derivative works, such
modified software should be clearly marked, so as not to confuse it with the
version available from LANL.

Additionally, this program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or (at your option) any
later version. Accordingly, this program is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
License for more details.
"""

import argparse as ap, textwrap as tw
import sys, os, time, subprocess
from gottcha.scripts import taxonomy as gt
import pandas as pd
import gc
from re import search,findall
from re import compile as recompile
from multiprocessing import Pool
from itertools import chain
import math
import logging

logger = logging.getLogger()

def parse_params( ver, args ):
    p = ap.ArgumentParser( prog='gottcha2.py', description="""Genomic Origin Through Taxonomic CHAllenge (GOTTCHA) is an
            annotation-independent and signature-based metagenomic taxonomic profiling tool
            that has significantly smaller FDR than other profiling tools. This program
            is a wrapper to map input reads to pre-computed signature databases using minimap2
            and/or to profile mapped reads in SAM format. (VERSION: %s)""" % ver)

    eg = p.add_mutually_exclusive_group( required=True )

    eg.add_argument( '-i','--input', metavar='[FASTQ]', nargs='+', type=ap.FileType('r'),
                    help="Input one or multiple FASTQ/FASTA file(s). Use space to separate multiple input files.")

    eg.add_argument( '-s','--sam', metavar='[SAMFILE]', nargs=1, type=ap.FileType('r'),
                    help="Specify the input SAM file. Use '-' for standard input.")

    p.add_argument( '-d','--database', metavar='[MINIMAP2_INDEX]', type=str, default=None,
                    help="The path of signature database. The database can be in FASTA format or minimap2 index (5 files).")

    p.add_argument( '-l','--dbLevel', metavar='[LEVEL]', type=str, default='',
                    choices=['superkingdom', 'phylum', 'class', 'order', 'family', 'genus', 'species', 'strain'],
                    help="""Specify the taxonomic level of the input database. You can choose one rank from "superkingdom", "phylum", "class", "order", "family", "genus", "species" and "strain". The value will be auto-detected if the input database ended with levels (e.g. GOTTCHA_db.species).""")

    p.add_argument( '-ti','--taxInfo', metavar='[FILE]', type=str, default='',
                    help="""Specify the path of taxonomy information directory (taxonomy_db). GOTTCHA2 will try to locate this file when user doesn't specify a path. If '--database' option is used, the program will try to find this file in the directory of specified database. If not, the 'database' directory under the location of gottcha.py will be used as default.""")

    p.add_argument( '-np','--nanopore', action="store_true",
                    help="Adjust options for Nanopore reads. The 'mismatch' option will be ignored. [-xm map-ont -mr 1]")

    p.add_argument( '-pm','--mismatch', metavar='<INT>', type=int, default=10,
                    help="Mismatch penalty for the aligner. [default: 10]")

    p.add_argument( '-e','--extract', metavar='[TAXID]', type=str, default=None,
                    help="""Extract reads mapping to a specific TAXID. [default: None]""" )

    p.add_argument( '-fm','--format', metavar='[STR]', type=str, default='tsv',
                    choices=['tsv','csv','biom'],
                    help='Format of the results; available options include tsv, csv or biom. [default: tsv]')

    p.add_argument( '-r','--relAbu', metavar='[FIELD]', type=str, default='ROLLUP_DOC',
                    choices=['ROLLUP_DOC','READ_COUNT','TOTAL_BP_MAPPED'],
                    help='The field will be used to calculate relative abundance. You can specify one of the following fields: "LINEAR_LENGTH", "TOTAL_BP_MAPPED", "READ_COUNT" and "LINEAR_DOC". [default: ROLLUP_DOC]')

    p.add_argument( '-t','--threads', metavar='<INT>', type=int, default=1,
                    help="Number of threads [default: 1]")

    p.add_argument( '-o','--outdir', metavar='[DIR]', type=str, default='.',
                    help="Output directory [default: .]")

    p.add_argument( '-p','--prefix', metavar='<STR>', type=str, required=False,
                    help="Prefix of the output file [default: <INPUT_FILE_PREFIX>]")

    p.add_argument( '-xm','--presetx', metavar='<STR>', type=str, required=False, default='sr',
                    choices=['sr','map-pb','map-ont'],
                    help="The preset option (-x) for minimap2. Default value 'sr' for short reads. [default: sr]")

    p.add_argument( '-mc','--minCov', metavar='<FLOAT>', type=float, default=0.005,
                    help="Minimum linear coverage to be considered valid in abundance calculation [default: 0.005]")

    p.add_argument( '-mr','--minReads', metavar='<INT>', type=int, default=3,
                    help="Minimum number of reads to be considered valid in abundance calculation [default: 3]")

    p.add_argument( '-ml','--minLen', metavar='<INT>', type=int, default=60,
                    help="Minimum unique length to be considered valid in abundance calculation [default: 60]")

    p.add_argument( '-mz','--maxZscore', metavar='<FLOAT>', type=float, default=10,
                    help="Maximum estimated zscore of depths of mapped region [default: 10]")

    p.add_argument( '-nc','--noCutoff', action="store_true",
                    help="Remove all cutoffs. This option is equivalent to use [-mc 0 -mr 0 -ml 0].")

    p.add_argument( '-c','--stdout', action="store_true",
                    help="Write on standard output.")

    eg.add_argument( '-v','--version', action="store_true",
                    help="Print version number.")

    p.add_argument( '--silent', action="store_true",
                    help="Disable all messages.")

    p.add_argument( '--debug', action="store_true",
                    help="Debug mode. Provide verbose running messages and keep all temporary files.")

    args_parsed = p.parse_args(args)

    """
    Checking options
    """
    if args_parsed.version:
        print( ver )
        sys.exit(0)

    if not args_parsed.database:
        p.error( '--database option is missing.' )

    if args_parsed.input and args_parsed.sam:
        p.error( '--input and --same are incompatible options.' )

    if args_parsed.database:
        #assign default path for database name
        if "/" not in args_parsed.database and not os.path.isfile( args_parsed.database + ".mmi" ):
            bin_dir = os.path.dirname(os.path.realpath(__file__))
            args_parsed.database = bin_dir + "/database/" + args_parsed.database

    if args_parsed.database and args_parsed.database.endswith(".mmi"):
        args_parsed.database.replace('.mmi','')

    if args_parsed.database and args_parsed.input:
        if not os.path.isfile( args_parsed.database + ".mmi" ):
            p.error( 'Database index %s.mmi not found.' % args_parsed.database )

    if not args_parsed.taxInfo:
        if args_parsed.database:
            db_dir = search(r'^(.*?)[^\/]+$', args_parsed.database )
            args_parsed.taxInfo = db_dir.group(1) + "/taxonomy_db"

            if not os.path.isdir(args_parsed.taxInfo):
                bin_dir = os.path.dirname(os.path.realpath(__file__))
                args_parsed.taxInfo = bin_dir + "/database"

            if not os.path.isdir(args_parsed.taxInfo):
                args_parsed.taxInfo = db_dir.group(1)

    if not args_parsed.prefix:
        if args_parsed.input:
            name = search(r'([^\/\.]+)\..*$', args_parsed.input[0].name )
            args_parsed.prefix = name.group(1)
        elif args_parsed.sam:
            name = search(r'([^\/]+).\w+.\w+$', args_parsed.sam[0].name )
            args_parsed.prefix = name.group(1)
        else:
            args_parsed.prefix = "GOTTCHA_"

    if not args_parsed.dbLevel:
        if args_parsed.database:
            major_ranks = {"superkingdom":1,"phylum":2,"class":3,"order":4,"family":5,"genus":6,"species":7, "strain":8}
            parts = args_parsed.database.split('.')
            for part in parts:
                if part in major_ranks:
                    args_parsed.dbLevel = part
                    break
        elif args_parsed.sam:
            name = search(r'\.gottcha_(\w+).sam$', args_parsed.sam[0].name )
            args_parsed.dbLevel = name.group(1)
        else:
            p.error( '--dbLevel is missing and cannot be auto-detected.' )

    if args_parsed.noCutoff:
        args_parsed.minCov = 0
        args_parsed.minReads = 0
        args_parsed.minLen = 0
        args_parsed.maxZscore = 99

    if args_parsed.nanopore:
        args_parsed.presetx = 'map-ont'
        args_parsed.minReads = 1

    return args_parsed

def dependency_check(cmd):
    try:
        subprocess.check_call([cmd, "--help"], stdout=subprocess.DEVNULL)
    except Exception as e:
        sys.stderr.write(f"[ERROR] {cmd}: {e}\n")
        sys.exit(1)

def worker(filename, chunkStart, chunkSize):
    """Make a dict out of the parsed, supplied lines"""
    # processing alignments in SAM format
    f = open( filename )
    f.seek(chunkStart)
    lines = f.read(chunkSize).splitlines()
    res={}

    for line in lines:
        k, r, m, n, rd, rs, rq, flag, cigr, pri_aln_flag, valid_flag = parse(line)
        if valid_flag:
            if k in res:
                res[k]["ML"] = res[k]["ML"] | m
                if pri_aln_flag and valid_flag:
                    res[k]["MB"] += r[1] - r[0] + 1
                    res[k]["MR"] += 1
                    res[k]["NM"] += n
            else:
                res[k]={}
                res[k]["ML"] = m
                if pri_aln_flag and valid_flag:
                    res[k]["MB"] = r[1] - r[0] + 1
                    res[k]["MR"] = 1
                    res[k]["NM"] = n
                else:
                    res[k]["MB"] = 0
                    res[k]["MR"] = 0
                    res[k]["NM"] = 0
    return res

def parse(line):
    """
    Parse SAM format
    read1   0   test    11  0   5S10M3S *   0   0   GGGGGCCCCCCCCCCGGG  HHHHHHHHHHHHHHHHHH  NM:i:0  MD:Z:10 AS:i:10 XS:i:0
    read2   16  test    11  0   3S10M5S *   0   0   GGGCCCCCCCCCCGGGGG  HHHHHHHHHHHHHHHHHH  NM:i:0  MD:Z:10 AS:i:10 XS:i:0
    """
    temp = line.split('\t')

    # adding mate number to the read name, first in pair (0x40) or second in pair (0x80)
    sam_flag = int(temp[1])
    if sam_flag & 64:
        temp[0] += ".1"
    elif sam_flag & 128:
        temp[0] += ".2"

    name = temp[0]
    match_len    = search(r'(\d+)M', temp[5])
    mismatch_len = search(r'NM:i:(\d+)', line)
    start = int(temp[3])
    end   = start + int(match_len.group(1)) - 1

    ref = temp[2].rstrip('|')
    ref = ref[: -2 if ref.endswith(".0") else None ]

    (acc, rstart, rend, taxid) = ref.split('|')
    rlen = int(rend)-int(rstart)+1
    mask = int( "%s%s"%("1"*(end-start+1), "0"*(rlen-end)), 2)

    primary_alignment_flag=False if int(temp[1]) & 256 else True
    valid_flag=True if (int(match_len.group(1)) >= rlen*0.5) or (int(match_len.group(1)) >= len(temp[9])*0.5) else False

    return ref, [start, end], mask, int(mismatch_len.group(1)), name, temp[9], temp[10], temp[1], temp[5], primary_alignment_flag, valid_flag

def time_spend( start ):
    done = time.time()
    elapsed = done - start
    return time.strftime( "%H:%M:%S", time.gmtime(elapsed) )

def chunkify(fname, size=1*1024*1024):
    fileEnd = os.path.getsize(fname)
    with open(fname, "rb") as f:
        chunkEnd = f.tell()
        while True:
            chunkStart = chunkEnd
            f.seek(size, 1)
            f.readline()
            # put all alignments of a read in the same chunck
            line = f.readline().decode('ascii')
            tmp = line.split('\t')
            if chunkEnd <= fileEnd and line:
                last = f.tell()
                line = f.readline().decode('ascii')
                while line.startswith(tmp[0]):
                    last = f.tell()
                    line = f.readline().decode('ascii')
                f.seek(last)
            # current position
            chunkEnd = f.tell()
            yield chunkStart, chunkEnd - chunkStart
            if chunkEnd > fileEnd:
                break

def process_sam_file( sam_fn, numthreads, numlines ):
    result = gt._autoVivification()
    mapped_reads = 0

    #clean memory
    gc.collect()

    print_message( "Parsing SAM files with %s subprocesses..."%numthreads, argvs.silent, begin_t, logfile )
    pool = Pool(processes=numthreads)
    jobs = []
    results = []

    for chunkStart,chunkSize in chunkify(sam_fn):
        jobs.append( pool.apply_async(worker, (sam_fn,chunkStart,chunkSize)) )

    #wait for all jobs to finish
    tol_jobs = len(jobs)
    cnt=0
    for job in jobs:
        results.append( job.get() )
        cnt+=1
        if argvs.debug: print_message( "[DEBUG] Progress: %s/%s (%.1f%%) chunks done."%(cnt, tol_jobs, cnt/tol_jobs*100), argvs.silent, begin_t, logfile )

    #clean up
    pool.close()

    print_message( "Merging results...", argvs.silent, begin_t, logfile )
    for res in results:
        for k in res:
            if k in result:
                result[k]["ML"] = result[k]["ML"] | res[k]["ML"]
                result[k]["MB"] += res[k]["MB"]
                result[k]["MR"] += res[k]["MR"]
                result[k]["NM"] += res[k]["NM"]
            else:
                result[k]={}
                result[k].update(res[k])

    # convert mapped regions to linear length
    refs = result.keys()
    for k in list(refs):
        if not result[k]["MR"]:
            del result[k]
        else:
            result[k]["LL"] = 0
            mapped_reads += result[k]["MR"]
            mask = result[k]["ML"]
            del result[k]["ML"]
            bitstr = bin(mask)
            result[k]["LL"] = len(bitstr[2:].replace("0",""))

    return result, mapped_reads

def is_descendant( taxid, taxid_ant ):
    fullLineage = gt.taxid2fullLineage( taxid )
    if "|%s|" % taxid_ant in fullLineage:
        return True
    else:
        return False

def extract_read_from_sam( sam_fn, o, taxid, numthreads ):
    pool = Pool(processes=numthreads)
    jobs = []

    for chunkStart,chunkSize in chunkify(sam_fn):
        jobs.append( pool.apply_async(ReadExtractWorker, (sam_fn,chunkStart,chunkSize,taxid)) )

    #wait for all jobs to finish
    for job in jobs:
        outread = job.get()
        o.write(outread)
        o.flush()

    #clean up
    pool.close()

def ReadExtractWorker( filename, chunkStart, chunkSize, taxid ):
    # output
    readstr=""
    # processing alignments in SAM format
    f = open( filename )
    f.seek(chunkStart)
    lines = f.read(chunkSize).splitlines()
    for line in lines:
        ref, region, mask, nm, rname, rseq, rq, flag, cigr, pri_aln_flag, valid_flag = parse(line)

        if not (pri_aln_flag and valid_flag): continue

        acc, start, stop, t = ref.split('|')
        fullLineage = gt.taxid2fullLineage(t)

        if int(flag) & 16:
            g = findall(r'\d+\w', cigr)
            cigr = "".join(list(reversed(g)))
            rseq = seqReverseComplement(rseq)
            rq = rq[::-1]

        if is_descendant( t, taxid ):
            readstr += "@%s %s:%s..%s %s\n%s\n+\n%s\n" % (rname, ref, region[0], region[1], cigr, rseq, rq)
    return readstr

def seqReverseComplement( seq ):
    seq1 = 'ACGTURYSWKMBDHVNTGCAAYRSWMKVHDBNacgturyswkmbdhvntgcaayrswmkvhdbn'
    seq_dict = { seq1[i]:seq1[i+16] for i in range(64) if i < 16 or 32<=i<48 }
    return "".join([seq_dict[base] for base in reversed(seq)])

def group_refs_to_strains(r):
    # covert mapping info to df
    r_df = pd.DataFrame.from_dict(r, orient='index').reset_index()
    r_df.rename(columns={"index": "RNAME"}, inplace=True)
    # retrieve sig fragment info
    r_df['RNAME'] = r_df['RNAME'].str.rstrip('|')
    r_df[['ACC','RSTART','REND','TAXID']] = r_df['RNAME'].str.split('|', expand=True)
    r_df['RSTART'] = r_df['RSTART'].astype(int)
    r_df['REND'] = r_df['REND'].astype(int)
    r_df['RLEN'] = r_df['REND']-r_df['RSTART']+1

    # group by strain
    str_df = r_df.groupby(['TAXID']).agg({
        'MB':'sum', # of mapped bases
        'MR':'sum', # of mapped reads
        'NM':'sum', # of mismatches
        'LL':'sum', # linear length
        'RLEN':'sum' # length of this signature fragments (mapped)
    }).reset_index()
    # total length of signatures
    str_df['TS'] = str_df['TAXID'].apply(lambda x: db_stats[x])
    str_df['bDOC'] = str_df['MB']/str_df['TS'] # bDOC: best Depth of Coverage of a strain
    str_df['bLC'] = str_df['LL']/str_df['TS'] # bLC:  best linear coverage of a strain
    str_df['RD'] = str_df['MB']/str_df['TS'] # roll-up DoC

    # rename columns
    str_df.rename(columns={
        "MB":  "TOTAL_BP_MAPPED",
        "MR":  "READ_COUNT",
        "NM":  "TOTAL_BP_MISMATCH",
        "LL":  "LINEAR_LEN",
        "RLEN":"MAPPED_SIG_LENGTH",
        "TS":  "TOL_SIG_LENGTH",
        "RD":  "ROLLUP_DOC",
        "bDOC":"BEST_DOC",
        "bLC": "BEST_LINEAR_COV"
    }, inplace=True)

    str_df['ZSCORE'] = str_df.apply(lambda x: pile_lvl_zscore(x.TOTAL_BP_MAPPED, x.TOL_SIG_LENGTH, x.LINEAR_LEN), axis=1)

    return str_df

def roll_up_taxonomy(r, db_stats, abu_col, tg_rank, mc, mr, ml, mz):
    """
    Take parsed SAM output and rollup to superkingdoms
    """
    major_ranks = {"superkingdom":1,"phylum":2,"class":3,"order":4,"family":5,"genus":6,"species":7,"strain":8}

    # roll up references to strains
    str_df = group_refs_to_strains(r)
    # produce columns for the final report at each ranks
    rep_df = pd.DataFrame()

    # qualified strain
    qualified_idx = (str_df['LINEAR_LEN']/str_df['TOL_SIG_LENGTH'] >= mc) & \
                    (str_df['READ_COUNT'] >= mr) & \
                    (str_df['LINEAR_LEN'] >= ml) & \
                    (str_df['ZSCORE'] <= mz)

    for rank in sorted(major_ranks, key=major_ranks.__getitem__):
        str_df['LVL_NAME'] = str_df['TAXID'].apply(lambda x: gt.taxid2lineageDICT(x, True, True)[rank]['name'])
        str_df['LVL_TAXID'] = str_df['TAXID'].apply(lambda x: gt.taxid2lineageDICT(x, True, True)[rank]['taxid'])
        str_df['LEVEL'] = rank

        logger.debug( str_df )

        # rollup strains that make cutoffs
        lvl_df = pd.DataFrame()
        if rank == 'strain':
            lvl_df = str_df
        else:
            lvl_df = str_df[qualified_idx].groupby(['LVL_NAME']).agg({
                'LEVEL':'first',
                'LVL_TAXID':'first',
                'TOTAL_BP_MAPPED': 'sum', 'READ_COUNT': 'sum', 'TOTAL_BP_MISMATCH': 'sum',
                'LINEAR_LEN': 'sum', 'MAPPED_SIG_LENGTH': 'sum', 'TOL_SIG_LENGTH': 'sum',
                'ROLLUP_DOC': 'sum', 'BEST_DOC': 'max', 'BEST_LINEAR_COV': 'max', 'ZSCORE': 'min',
            }).reset_index().copy()

        lvl_df['ABUNDANCE'] = lvl_df[abu_col]
        tol_abu = lvl_df[abu_col].sum()
        lvl_df['REL_ABUNDANCE'] = lvl_df[abu_col]/tol_abu

        # add NOTE if ranks is higher than target rank
        lvl_df['NOTE'] = ""
        if major_ranks[rank] > major_ranks[tg_rank]:
            lvl_df['NOTE'] = f"Not shown ({rank}-result biased); "

        # concart ranks-dataframe to the report-dataframe
        rep_df = pd.concat([rep_df, lvl_df.sort_values('ABUNDANCE', ascending=False)], sort=False)

    rep_df["LINEAR_COV"] = rep_df["LINEAR_LEN"]/rep_df["TOL_SIG_LENGTH"]
    rep_df["LINEAR_COV_MAPPED_SIG"] = rep_df["LINEAR_LEN"]/rep_df["MAPPED_SIG_LENGTH"]
    rep_df["DOC"] = rep_df["TOTAL_BP_MAPPED"]/rep_df["TOL_SIG_LENGTH"]
    rep_df["LINEAR_DOC"] = rep_df["TOTAL_BP_MAPPED"]/rep_df["LINEAR_LEN"]

    # add filtered reason to NOTE column
    filtered = (rep_df['LINEAR_COV'] < mc)
    rep_df.loc[filtered, 'NOTE'] += "Filtered out (minCov > " + rep_df.loc[filtered, 'LINEAR_COV'].astype(str) + "); "
    filtered = (rep_df['READ_COUNT'] < mr)
    rep_df.loc[filtered, 'NOTE'] += "Filtered out (minReads > " + rep_df.loc[filtered, 'READ_COUNT'].astype(str) + "); "
    filtered = (rep_df['LINEAR_LEN'] < ml)
    rep_df.loc[filtered, 'NOTE'] += "Filtered out (minLen > " + rep_df.loc[filtered, 'LINEAR_LEN'].astype(str) + "); "
    filtered = (rep_df['ZSCORE'] > mz)
    rep_df.loc[filtered, 'NOTE'] += "Filtered out (maxZscore < " + rep_df.loc[filtered, 'ZSCORE'].astype(str) + "); "

    rep_df.drop(columns=['TAXID'], inplace=True)
    rep_df.rename(columns={"LVL_NAME": "NAME", "LVL_TAXID": "TAXID"}, inplace=True)

    return rep_df

def pile_lvl_zscore(tol_bp, genome_size, linear_len):
    try:
        doc = tol_bp/genome_size
        lin_doc = tol_bp/linear_len
        v = (linear_len*(lin_doc-doc)**2 + (genome_size-linear_len)*(doc)**2)/genome_size
        sd = math.sqrt(v)
        return (lin_doc-doc)/sd
    except:
        return 99

def generaete_taxonomy_file(rep_df, o, fullreport_o, fmt="tsv"):
    """
    output result in tsv or csv format
    """
    # Fields for full mode
    cols = ['LEVEL', 'NAME', 'TAXID', 'READ_COUNT', 'TOTAL_BP_MAPPED',
            'TOTAL_BP_MISMATCH', 'LINEAR_LEN', 'LINEAR_DOC', 'ROLLUP_DOC', 'REL_ABUNDANCE',
            'LINEAR_COV', 'LINEAR_COV_MAPPED_SIG', 'BEST_LINEAR_COV', 'MAPPED_SIG_LENGTH', 'TOL_SIG_LENGTH',
            'ABUNDANCE', 'ZSCORE', 'NOTE']

    qualified_idx = (rep_df['NOTE']=="")
    qualified_df = rep_df.loc[qualified_idx, cols[:10]]

    sep = ',' if fmt=='csv' else '\t'
    # save full report
    rep_df[cols].to_csv(fullreport_o, index=False, sep=sep, float_format='%.6f')
    # save summary
    qualified_df.to_csv(o, index=False, sep=sep, float_format='%.6f')

    return True

def generaete_biom_file(res_df, o, tg_rank, sampleid):
    """
    output result in biom format
    """
    import numpy as np
    import biom
    from biom.table import Table
    if biom.__version__ < '2.1.7':
        sys.exit("[ERROR] Biom library requires v2.1.7 or above.\n")

    target_df = pd.DataFrame()
    target_idx = (res_df['LEVEL']==tg_rank)
    target_df = res_df.loc[target_idx, ['ABUNDANCE','TAXID']]
    target_df['LINEAGE'] = target_df['TAXID'].apply(lambda x: gt.taxid2lineage(x, True, True)).str.split('|')

    sample_ids = [sampleid]
    data = np.array(target_df['ABUNDANCE']).reshape(len(target_df), 1)
    observ_ids = target_df['TAXID']
    observ_metadata = [{'taxonomy': x} for x in target_df['LINEAGE'].tolist()]
    biom_table = Table(data, observ_ids, sample_ids, observ_metadata, table_id='GOTTCHA2')
    biom_table.to_json('GOTTCHA2', direct_io=o)

    return True

def generaete_lineage_file(target_df, o, tg_rank):
    """
    output abundance-lineage in tsv format
    """
    lineage_df = target_df['TAXID'].apply(lambda x: gt.taxid2lineage(x, True, True)).str.split('|', expand=True)
    result = pd.concat([target_df['ABUNDANCE'], lineage_df], axis=1, sort=False)
    result.to_csv(o, index=False, header=False, sep='\t', float_format='%.4f')

    return True

def readMapping(reads, db, threads, mm_penalty, presetx, samfile, logfile, nanopore):
    """
    mapping reads to database
    """
    input_file = " ".join([x.name for x in reads])

    sr_opts = f"-x {presetx} -k24 -A1 -B{mm_penalty} -O30 -E30 -a -N20 --secondary=no -n1 -p1 -m24 -s30"
    if nanopore:
        sr_opts = f"-x {presetx} -N20 --secondary=no -a"

    bash_cmd   = "set -o pipefail; set -x;"
    mm2_cmd    = f"minimap2 {sr_opts} -t{threads} {db}.mmi {input_file}"
    filter_cmd = "samtools view -F4 -F2048"
    cmd        = "%s %s 2>> %s | %s > %s"%(bash_cmd, mm2_cmd, logfile, filter_cmd, samfile)

    proc = subprocess.Popen( cmd, shell=True, executable='/bin/bash', stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    outs, errs = proc.communicate()
    exitcode = proc.poll()

    return exitcode, mm2_cmd, errs

def loadDatabaseStats(db_stats_file):
    """
    loading database stats from db_path.stats

    The input stats file is a 8 column tab delimited text file:
    1. Rank
    2. Name
    3. Taxid
    4. Superkingdom
    5. NumOfSeq
    6. Max
    7. Min
    8. TotalLength

    We will only save stain level info with their taxid (2) and total signature length (8).
    """
    db_stats = {}

    with open(db_stats_file) as f:
        for line in f:
            fields = line.split("\t")
            major_ranks = {"superkingdom":1,"phylum":2,"class":3,"order":4,"family":5,"genus":6,"species":7, "strain":8}
            if fields[0] in major_ranks:
                db_stats[fields[2]] = int(fields[7])
            else:
                continue

    return db_stats

def print_message(msg, silent, start, logfile, errorout=0):
    message = "[%s] %s\n" % (time_spend(start), msg)

    with open( logfile, "a" ) as f:
        f.write( message )
        f.close()

    if errorout:
        sys.exit( message )
    elif not silent:
        sys.stderr.write( message )

def main(args):
    global argvs
    global logfile
    global begin_t
    global db_stats
    argvs = parse_params( __version__, args )
    begin_t  = time.time()
    sam_fp   = argvs.sam[0] if argvs.sam else ""
    samfile  = "%s/%s.gottcha_%s.sam" % ( argvs.outdir, argvs.prefix, argvs.dbLevel ) if not argvs.sam else sam_fp.name
    logfile  = "%s/%s.gottcha_%s.log" % ( argvs.outdir, argvs.prefix, argvs.dbLevel )
    lines_per_process = 10000

    if argvs.debug:
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(asctime)s [%(levelname)s] %(module)s: %(message)s',
            datefmt='%Y-%m-%d %H:%M',
        )
        logger = logging.getLogger()

    #dependency check
    if sys.version_info < (3,4):
        sys.exit("[ERROR] Python 3.4 or above is required.")
    dependency_check("minimap2")
    dependency_check("gawk")

    #prepare output object
    argvs.relAbu = argvs.relAbu.upper()
    outfile_full = "%s/%s.full.tsv" % (argvs.outdir, argvs.prefix)
    outfile_lineage = "%s/%s.lineage.tsv" % (argvs.outdir, argvs.prefix)

    # remove previous log file if exists
    if os.path.isfile(logfile):
        os.remove(logfile)

    out_fp = sys.stdout
    outfile = "STDOUT"

    if not argvs.stdout:
        #create output directory if not exists
        if not os.path.exists(argvs.outdir):
            os.makedirs(argvs.outdir)
        ext = "tsv"
        outfile = "%s/%s.tsv" % (argvs.outdir, argvs.prefix)
        if argvs.extract:
            outfile = "%s/%s.extract%s.fastq" % (argvs.outdir, argvs.prefix, argvs.extract)
        elif argvs.format == "csv":
            outfile = "%s/%s.csv" % (argvs.outdir, argvs.prefix)
        elif argvs.format == "biom":
            outfile = "%s/%s.biom" % (argvs.outdir, argvs.prefix)

        out_fp = open(outfile, 'w')

    print_message( "Starting GOTTCHA (v%s)" % __version__, argvs.silent, begin_t, logfile )
    print_message( "Arguments and dependencies checked:", argvs.silent, begin_t, logfile )
    if argvs.input:
        print_message( "    Input reads      : %s" % [x.name for x in argvs.input],     argvs.silent, begin_t, logfile )
    print_message( "    Input SAM file   : %s" % samfile,         argvs.silent, begin_t, logfile )
    print_message( "    Database         : %s" % argvs.database,  argvs.silent, begin_t, logfile )
    print_message( "    Database level   : %s" % argvs.dbLevel,   argvs.silent, begin_t, logfile )
    print_message( "    Mismatch penalty : %s" % argvs.mismatch,  argvs.silent, begin_t, logfile )
    print_message( "    Abundance        : %s" % argvs.relAbu,    argvs.silent, begin_t, logfile )
    print_message( "    Output path      : %s" % argvs.outdir,    argvs.silent, begin_t, logfile )
    print_message( "    Prefix           : %s" % argvs.prefix,    argvs.silent, begin_t, logfile )
    print_message( "    Extract taxid    : %s" % argvs.extract,   argvs.silent, begin_t, logfile )
    print_message( "    Threads          : %d" % argvs.threads,   argvs.silent, begin_t, logfile )
    print_message( "    Minimal L_DOC    : %s" % argvs.minCov,    argvs.silent, begin_t, logfile )
    print_message( "    Minimal L_LEN    : %s" % argvs.minLen,    argvs.silent, begin_t, logfile )
    print_message( "    Minimal reads    : %s" % argvs.minReads,  argvs.silent, begin_t, logfile )

    #load taxonomy
    print_message( "Loading taxonomy information...", argvs.silent, begin_t, logfile )
    custom_taxa_tsv = None
    if os.path.isfile( argvs.database + ".tax.tsv" ):
        custom_taxa_tsv = argvs.database+".tax.tsv"
    gt.loadTaxonomy( argvs.taxInfo, custom_taxa_tsv )
    print_message( "Done.", argvs.silent, begin_t, logfile )

    #load database stats
    print_message( "Loading database stats...", argvs.silent, begin_t, logfile )
    if os.path.isfile( argvs.database + ".stats" ):
        db_stats = loadDatabaseStats(argvs.database+".stats")
    else:
        sys.exit( "[%s] ERROR: %s not found.\n" % (time_spend(begin_t), argvs.database+".stats") )
    print_message( "Done.", argvs.silent, begin_t, logfile )

    if argvs.input:
        print_message( "Running read-mapping...", argvs.silent, begin_t, logfile )
        exitcode, cmd, msg = readMapping( argvs.input, argvs.database, argvs.threads, argvs.mismatch, argvs.presetx, samfile, logfile, argvs.nanopore )
        print_message( "Logfile saved to %s." % logfile, argvs.silent, begin_t, logfile )
        print_message( "COMMAND: %s" % cmd, argvs.silent, begin_t, logfile )
        if exitcode != 0:
            sys.exit( "[%s] ERROR: error occurred while running read mapping (exit: %s, message: %s).\n" % (time_spend(begin_t), exitcode, msg) )
        else:
            print_message( "Done mapping reads to %s signature database." % argvs.dbLevel, argvs.silent, begin_t, logfile )
            print_message( "Mapped SAM file saved to %s." % samfile, argvs.silent, begin_t, logfile )
            sam_fp = open( samfile, "r" )

    if argvs.extract:
        extract_read_from_sam( os.path.abspath(samfile), out_fp, argvs.extract, argvs.threads )
        print_message( "Done extracting reads to %s." % outfile, argvs.silent, begin_t, logfile )
    else:
        (res, mapped_r_cnt) = process_sam_file( os.path.abspath(samfile), argvs.threads, lines_per_process)
        print_message( "Done processing SAM file. %s qualified mapped reads." % mapped_r_cnt, argvs.silent, begin_t, logfile )

        if mapped_r_cnt:
            res_df = roll_up_taxonomy(res, db_stats, argvs.relAbu, argvs.dbLevel , argvs.minCov, argvs.minReads, argvs.minLen, argvs.maxZscore)
            print_message( "Done taxonomy rolling up.", argvs.silent, begin_t, logfile )

            if not len(res_df):
                print_message( "No qualified taxonomy profiled.", argvs.silent, begin_t, logfile )
            else:
                # generate output files
                if argvs.format == "biom":
                    generaete_biom_file(res_df, out_fp, argvs.dbLevel, argvs.prefix)
                else:
                    generaete_taxonomy_file(res_df, out_fp, outfile_full, argvs.format)
                # generate lineage file
                target_idx = (res_df['LEVEL']==argvs.dbLevel)
                target_df = res_df.loc[target_idx, ['ABUNDANCE','TAXID']]
                tax_num = len(target_df)
                if tax_num:
                     generaete_lineage_file(target_df, outfile_lineage, argvs.dbLevel)

                print_message( "%s qualified %s profiled; Results saved to %s." %(tax_num, argvs.dbLevel, outfile), argvs.silent, begin_t, logfile )
        else:
            print_message( "GOTTCHA2 stopped.", argvs.silent, begin_t, logfile )

if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python

# Po-E (Paul) Li
# B-11, Los Alamos National Lab
# Date: 05/15/2016

import sys
import os
import tarfile
import logging
from typing import Union, Optional

try:
    from . import __version__
except:
    __version__ = 'standalone'

logger = logging.getLogger()

# Set default path of `taxonomy_db/` and `major_level_to_abbr.json`:
# The default `taxonomy_db/` path is your the location
lib_path = os.path.dirname(os.path.realpath(__file__))
taxonomy_dir = f"{lib_path}/taxonomy_db"
abbr_json_path = f"{taxonomy_dir}/major_level_to_abbr.json"

if os.path.isdir( "./taxonomy_db" ):
    taxonomy_dir = "./taxonomy_db"
    # if there is a new `major_level_to_abbr.json` in in the taxonomy_dir, use the json file.
    # Otherwise, use the default one comes with this package
    if os.path.isfile( f"{taxonomy_dir}/major_level_to_abbr.json" ):
        abbr_json_path = f"{taxonomy_dir}/major_level_to_abbr.json"
    else:
        pass
elif os.path.isdir( os.getcwd()+"/taxonomy_db" ):
    taxonomy_dir = os.getcwd()+"/taxonomy_db"

# init global dir
taxDepths      = {}
taxParents     = {}
taxRanks       = {}
taxNames       = {}
taxMerged      = {}
taxNumChilds   = {}
accTid         = {}
tidLineage     = {}
tidLineageDict = {}
nameTid        = {}
major_level_to_abbr = {}
abbr_to_major_level = {}
df_names = None

# --- helper functions ---
def _getTaxDepth(tid: str) -> str:
    """Get the depth of a taxonomy ID [warning: only support Kraken taxa inputs]"""
    if tid in taxMerged: tid = taxMerged[tid]
    return int(taxDepths[tid])

def _getTaxName(tid: str) -> str:
    """Get the name of a taxonomy ID"""
    if tid in taxMerged: tid = taxMerged[tid]
    return taxNames[tid]

def _getTaxParent(tid: str) -> str:
    """Get the parent taxid of a taxonomy ID"""
    if tid in taxMerged: tid = taxMerged[tid]
    return taxParents[tid]

def _getTaxRank(tid: str) -> str:
    """Get the rank of a taxonomy ID"""
    if tid in taxMerged: tid = taxMerged[tid]
    return taxRanks[tid]

def _die(msg: str) -> str:
    sys.exit(msg)

def _checkTaxonomy(tid: Union[int, str]):
    """Check if a taxonomy ID is present in the taxonomy database"""

    if not len(taxParents):
        logger.fatal("Taxonomy not loaded. \"loadTaxonomy()\" must be called first.")
        _die("Taxonomy not loaded. \"loadTaxonomy()\" must be called first.")

    if tid:
        # tid must be in string type
        tid = str(tid)

        # convert to merged tid first if needs
        if tid in taxMerged:
            logger.info( f"Merged tid found: {tid} -> {taxMerged[tid]}." )
            tid = taxMerged[tid]

        if (tid in taxNames) and (tid in taxParents):
            return tid
        else:
            return "unknown"

def _taxid2fullLink(tid: Union[int, str]) -> dict:
    """
    Returns a dictionary containing the full lineage of the target taxon (e.g. {`[parent_tid]`: `tid`,...}).

    Args:
        tid (Union[int, str]): Taxonomy ID of the target taxon.

    Returns:
        dict: A dictionary containing the full lineage of the target taxon.

    """
    tid = _checkTaxonomy(tid)
    if tid == "unknown": return {}
    link = _autoVivification()

    while tid != '1':
        name = _getTaxName(tid)
        if not name: break
        
        parID = _getTaxParent(tid)
        link[parID] = tid
        tid = parID
    
    return link


def _taxid2lineage(tid: Union[int, str], 
                   all_major_rank: bool=True, 
                   print_strain: bool=True, 
                   space2underscore: bool=False, 
                   guess_type: bool=False):
    """
    Given a taxonomic ID, returns the lineage of the taxon as a dictionary.
    The lineage is represented as a dictionary where the keys are the taxonomic ranks and the values are the corresponding taxon names.
    If the taxon is not found in the nodes dictionary, an empty dictionary is returned.
    """
    tid_orig = tid
    tid = _checkTaxonomy( tid )
    if tid == "unknown": 
        logger.debug( f"Unknown taxid: {tid_orig}" )
        return {}
    # if tid in tidLineageDict: return tidLineageDict[tid]

    info = _autoVivification()
    level = {abbr: '' for abbr in abbr_to_major_level}

    rank = taxid2rank(tid)
    orig_rank = rank
    name = taxid2name(tid)
    str_name = name
    if space2underscore: str_name = str_name.replace(" ", "_")

    logger.debug( f"orig_rank: {orig_rank}, str_name: {str_name}" )

    while tid:
        if rank in major_level_to_abbr:
            if space2underscore: name = name.replace(" ", "_")
            level[major_level_to_abbr[rank]] = name

            #for output JSON
            info[rank]["name"] = name
            info[rank]["taxid"] = tid

        tid = _getTaxParent(tid)
        rank = _getTaxRank(tid)
        name = _getTaxName(tid)

        if name == 'root': break

    # try to get the closest "no_rank" taxa to "type" representing subtype/group (mainly for virus)
    if guess_type==True:
        typeTID = taxid2type(tid)
        if typeTID:
            info["type"]["name"]  = _getTaxName(typeTID)
            info["type"]["taxid"] = typeTID

    last = str_name

    ranks = list(abbr_to_major_level.keys())
    ranks.reverse()
    idx = 0
    
    # input taxid is a major rank
    if orig_rank in major_level_to_abbr:
        idx = ranks.index( major_level_to_abbr[orig_rank] )
    # if not, find the next major rank
    else:
        nmtid = taxid2nearestMajorTaxid( tid )
        nmrank = taxid2rank( nmtid )
        if nmrank == "root":
            idx = 7
        else:
            idx = ranks.index( major_level_to_abbr[nmrank] )

    for lvl in ranks[idx:]:
        if all_major_rank == False:
            if not level[lvl]: continue

        if not level[lvl]:
            level[lvl] = "%s - no_%s_rank"%(last,lvl)
            info[abbr_to_major_level[lvl]]["name"]  = "%s - no_%s_rank"%(last,lvl)
            info[abbr_to_major_level[lvl]]["taxid"] = 0

        last=level[lvl]

    if print_strain==True:
        if orig_rank == "strain":
            info["strain"]["name"]  = str_name
            info["strain"]["taxid"] = tid

    tidLineageDict[tid] = info
    return info

def _loadAbbrJson(abbr_json_path: str) -> None:
    """Load abbreviations for major ranks from a JSON file"""

    import json
    global major_level_to_abbr, abbr_to_major_level

    logging.debug(f'Loading major level to abbr from {abbr_json_path}...')

    if os.path.isfile(abbr_json_path):
        # Opening JSON file
        with open(abbr_json_path) as f:    
            major_level_to_abbr = json.load(f)
            f.close
    else:
        major_level_to_abbr = {
            "superkingdom" : "sk",
            "phylum"       : "p",
            "class"        : "c",
            "order"        : "o",
            "family"       : "f",
            "genus"        : "g",
            "species"      : "s",
            "strain"       : "n"
        }
        logger.info( f"None of the abbreviations for major ranks loaded from a JSON file. Default abbreviations use." )
        
    if len(major_level_to_abbr):
        abbr_to_major_level = {v: k for k, v in major_level_to_abbr.items()}
    else:
        logger.fatal( f"None of the major level to aberration loaded from {abbr_json_path}." )
        _die(f"[ERROR] None of the major level to aberration loaded from {abbr_json_path}.")

class _autoVivification(dict):
    """Implementation of perl's autovivification feature."""
    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value

# --- main functions ---

def taxid2rank(tid: Union[int, str], guess_strain: bool=True) -> str:
    """
    Get the taxonomic rank of a given taxonomic ID.
    
    Args:
        tid (Union[int, str]): Taxonomic ID.
        guess_strain (bool, optional): Whether to guess the strain when the rank, strain, is not available. 
            Defaults to True.
    
    Returns:
        str: The taxonomic rank.
    """

    tid = _checkTaxonomy(tid)
    if tid == "unknown": return "unknown"

    if tid == '1':
        return "root"

    if _getTaxRank(tid) == "no rank" and guess_strain:
        # a leaf taxonomy is a strain
        if taxidIsLeaf(tid):
            return "strain"
        # if not
        else:
            nmtid = taxid2nearestMajorTaxid(tid)
            nmrank = _getTaxRank(nmtid)
            if nmrank == "species":
                return "species - others"
            else:
                return "others"
    
    return _getTaxRank(tid)

def taxid2name(tid: Union[int, str]) -> str:
    """
    Get the taxonomic name of a given taxonomic ID.
    
    Args:
        tid (Union[int, str]): Taxonomic ID.
    
    Returns:
        str: The taxonomic name.
    """

    tid = _checkTaxonomy(tid)
    if tid == "unknown":
        return "unknown"
    else:
        return _getTaxName(tid)

def taxid2depth(tid: Union[int, str]) -> int:
    """
    Get the depth of a given taxonomic ID.
    
    Args:
        tid (Union[int, str]): Taxonomic ID.
    
    Returns:
        int: The depth of the taxonomic ID in the taxonomy tree.
    """

    tid = _checkTaxonomy(tid)
    if tid == "unknown":
        return "unknown"
    else:
        return _getTaxDepth(tid)

def taxid2type(tid: Union[int, str]):
    """
    Guessing the `type` of a given taxonomic ID. It could be rank of `group`, `serotype`...etc.
    
    Args:
        tid (Union[int, str]): Taxonomic ID.
    
    Returns:
        Union[int, str]: The type of the taxonomic ID.
    """

    tid = _checkTaxonomy(tid)
    if tid == "unknown": return "unknown"

    origID = tid
    lastID = tid
    tid = _getTaxParent(tid)

    while tid != '1' and _getTaxRank(tid) != 'species':
        lastID = tid
        tid = _getTaxParent(tid)

    if _getTaxRank(tid) != 'species':
        tid = 0
    else:
        tid = lastID
        if tid == origID: tid = 0

    return tid

def taxid2parent(tid: Union[int, str], norank: bool=False) -> str:
    """
    Get the parent of a given taxonomic ID.
    
    Args:
        tid (Union[int, str]): Taxonomic ID.
        norank (bool): Find the next parent if it's 'no-rank'.
    
    Returns:
        str: The taxonomic ID of the parent.
    """

    tid = _checkTaxonomy(tid)
    if tid == "unknown": return "unknown"

    tid = _getTaxParent(tid)

    if not norank:
        while tid != '1' and (_getTaxRank(tid) == 'no rank'):
            tid = _getTaxParent(tid)

    return tid

def name2taxid(name, rank=None, superkingdom=None, fuzzy=True, cutoff=0.7, max_matches=3, reset=False, expand=True) -> list:
    """
    Get the taxonomic ID of a given taxonomic name.
    
    Args:
        name (str): Taxonomic scientific name.
        rank (str, optional): The expected rank of the taxonomic name.
        superkingdom (str, optional): The expected superkingdom of the taxonomic name.
        fuzzy (bool, optional): Whether to allow fuzzy search. Defaults to True.
        cutoff (float, optional): Similarity cutoff for difflib.get_close_matches(). 
            Only apply to `expand` mode. Cutoff will set to 1 if `fuzzy` set to False. Defaults to 0.7.
        max_matches (int, optional): Reporting max number of taxid. Defaults to 3.
        reset (bool, optional): Mapping results are cached. Whether to clean up previous searches. 
            Defaults to False.
        expand (bool, optional): Search the entire 'names.dmp' if True, otherwise search sientific names only. 
            Defaults to False.
    
    Returns:
        list: The list of matched taxonomic ID.
    """
    global nameTid, df_names, taxonomy_dir
    import pandas as pd

    if not fuzzy: cutoff=1

    # if expand is True, loading names.dmp
    names_dmp_file = taxonomy_dir+"/names.dmp"

    if df_names is None and expand and os.path.isfile( names_dmp_file ):
        df_names = pd.read_csv(names_dmp_file, 
                         sep='\t|\t', 
                         engine='python', 
                         header=None, 
                         names=['taxid', 'name', 'annot', 'type'], 
                         usecols=['taxid','name'],
                         index_col='name')
    
    if not name in nameTid or reset:
        matched_taxid = []

        if expand:
            import difflib
            matches = difflib.get_close_matches(name, df_names.index, max_matches, cutoff)
            logger.debug(f'{name}: {matches}')
            df_temp = df_names.loc[matches,:]

            if rank:
                df_temp['rank'] = df_temp.taxid.apply(taxid2rank)
                idx = df_temp['rank']==rank
                df_temp = df_temp[idx]
            
            if superkingdom:
                df_temp['sk'] = df_temp.taxid.apply(lambda x: taxid2nameOnRank(x, 'superkingdom'))
                idx = df_temp['sk']==superkingdom
                df_temp = df_temp[idx]
            
            nameTid[name] = df_temp.head(max_matches).taxid.to_list()
            return nameTid[name]

        else: # searching scientific names only
            for taxid in taxNames:
                if fuzzy==True:
                    if not name in taxNames[taxid]:
                        continue
                else:
                    if name!=taxNames[taxid]:
                        continue
                
                if rank:
                    if _getTaxRank(taxid)==rank:
                        matched_taxid.append(taxid)
                else:
                    matched_taxid.append(taxid)

                # return when the first match found
                if len(matched_taxid)==max_matches:
                    nameTid[name] = matched_taxid
                    return nameTid[name]
            
            nameTid[name] = matched_taxid
            return nameTid[name][:max_matches]
    else:
        return nameTid[name][:max_matches]

def taxid2nameOnRank(tid: Union[int, str], target_rank=None) -> str:
    """
    Get the taxonomic name of a given taxonomic ID at a specific rank.
    
    Args:
        tid (Union[int, str]): Taxonomic ID.
        target_rank (str, optional): Target rank. Defaults to None.
    
    Returns:
        str: The taxonomic name at the target rank.
    """

    tid = _checkTaxonomy(tid)
    if tid == "unknown": return "unknown"
    if tid == 1: return "root"
    if target_rank == "root": return "root"

    if tid:
        rank = _getTaxRank(tid)
        name = _getTaxName(tid)

        if target_rank == "strain" and taxidIsLeaf(tid):
            return name

        while tid:
            if rank.upper() == target_rank.upper(): return name
            if name == 'root': break
            tid = _getTaxParent(tid)
            rank = _getTaxRank(tid)
            name = _getTaxName(tid)
    else:
        return ""

def taxid2taxidOnRank(tid: Union[int, str], target_rank=None ) -> str:
    """
    Returns the taxonomy ID of the nearest parent taxon at the specified rank.

    Args:
        tid (Union[int, str]): Taxonomy ID of the target taxon.
        target_rank (str): The target rank to search for. Defaults to None.

    Returns:
        str: Taxonomy ID of the nearest parent taxon at the specified rank.

    """

    tid = _checkTaxonomy(tid)
    if tid == "unknown": return "unknown"

    if tid:
        rank = _getTaxRank(tid)
        name = _getTaxName(tid)

        if target_rank == rank or ( target_rank == 'strain' and rank == 'no rank'): return tid
        if target_rank == "root": return 1

        while tid:
            if rank.upper() == target_rank.upper(): return tid
            if name == 'root': break

            tid = _getTaxParent(tid)
            rank = _getTaxRank(tid)
            name = _getTaxName(tid)
    else:
        return ""

def taxidIsLeaf(tid: Union[int, str]) -> bool:
    """
    Checks if the taxonomy ID corresponds to a leaf node in the taxonomic tree.

    Args:
        tid (Union[int, str]): Taxonomy ID of the target taxon.

    Returns:
        bool: True if the taxon is a leaf node, False otherwise.

    """

    tid = _checkTaxonomy(tid)
    if tid == "unknown": return False
    if not tid in taxNumChilds:
        return True
    else:
        return False


def taxid2fullLineage(tid: Union[int, str], sep: str='|', use_rank_abbr=False, space2underscore=True) -> str:
    """
    Returns the full lineage of the target taxon in a specified format.

    Args:
        tid (Union[int, str]): Taxonomy ID of the target taxon.
        sep (str): Separator used to separate fields in the output. Defaults to '|'.
        use_rank_abbr (bool): If True, abbreviated rank names are used. Defaults to False.
        space2underscore (bool): If True, spaces in the output are replaced by underscores. Defaults to True.

    Returns:
        str: Full lineage of the target taxon in the specified format.

    """
    link = _taxid2fullLink(tid)
    texts = []
    if len(link):
        for p_taxID in link:
            tid = link[p_taxID]
            rank = _getTaxRank(tid)
            name = _getTaxName(tid)

            if use_rank_abbr and (rank in major_level_to_abbr):
                rank =  major_level_to_abbr[rank]

            if sep == ';':
                texts.append(f"{rank}__{name}")
            else:
                texts.append(f"{rank}|{tid}|{name}")
    
    texts.reverse()

    if space2underscore:
        return sep.join(texts).replace(' ', '_')
    else:
        return sep.join(texts)

def taxid2fullLinkDict(tid: Union[int, str]) -> str:
    """
    Returns a dictionary containing the full lineage of the target taxon.

    Args:
        tid (Union[int, str]): Taxonomy ID of the target taxon.

    Returns:
        dict: A dictionary containing the full lineage of the target taxon.

    """
    return _taxid2fullLink(tid)

def taxid2nearestMajorTaxid(tid: Union[int, str]) -> str:
    """
    Returns the taxonomy ID of the nearest parent taxon at a major rank.

    Args:
        tid (Union[int, str]): Taxonomy ID of the target taxon.

    Returns:
        str: Taxonomy ID of the nearest parent taxon at a major rank.

    """

    tid = _checkTaxonomy(tid)
    if tid == "unknown": return "unknown"
    ptid = _getTaxParent(tid)
    while ptid != '1':
        tmp = _getTaxRank(ptid)
        if tmp in major_level_to_abbr:
            return ptid
        else:
            ptid = _getTaxParent(ptid)

    return "1"

def taxid2lineage(tid: Union[int, str], all_major_rank=True, print_strain=False, space2underscore=False, sep="|") -> str:
    """
    Returns the taxonomic lineage for a given taxonomic identifier (tid) as a formatted string.

    Parameters:
        tid (Union[int, str]): A taxonomic identifier, which can be either an integer or a string.
        all_major_rank (bool): If True, all major taxonomic ranks will be included in the lineage; if False, only the lowest common ancestor will be included. Default is True.
        print_strain (bool): If True, strain information will be included in the lineage; if False, it will be omitted. Default is False.
        space2underscore (bool): If True, spaces in the taxonomic names will be replaced with underscores; if False, they will be left as spaces. Default is False.
        sep (str): The separator used to join the taxonomic ranks, taxids, and names in the returned string. Default is "|".

    Returns:
        str: A formatted string containing the taxonomic lineage information, with each rank separated by the specified separator (default is "|").

    """
    
    lineage = _taxid2lineage( tid, all_major_rank, print_strain, space2underscore)
    texts = []
    for rank in major_level_to_abbr:
        if rank in lineage:
            if print_strain==False and rank=="strain":
                continue
            if sep == ";":
                texts.append( f"{major_level_to_abbr[rank]}__{lineage[rank]['name']}" )
            else:
                texts.append( f"{rank}|{lineage[rank]['taxid']}|{lineage[rank]['name']}" ) 
    
    if space2underscore:
        return sep.join(texts).replace(' ', '_')
    else:
        return sep.join(texts) 

def taxid2lineageDICT(tid: Union[int, str], all_major_rank=True, print_strain=True, space2underscore=False, guess_type=False):
    return _taxid2lineage( tid, all_major_rank, print_strain, space2underscore, guess_type)

def lca_taxid(taxids: list) -> str:
    """ lca_taxid
    Return lowest common ancestor (LCA) taxid of input taxids
    """
    ranks = ['strain','species','genus','family','order','class','phylum','superkingdom']

    merged_dict = _autoVivification()
    for tid in taxids:
        lineage = taxid2lineageDICT(tid)
        for r in ranks:
            if not r in lineage:
                ttid = "0"
            else:
                ttid = lineage[r]['taxid']

            if ttid in merged_dict[r]:
                merged_dict[r][ttid] += 1
            else:
                merged_dict[r][ttid] = 1

    for r in ranks:
        if len(merged_dict[r]) == 1:
            for ttid in merged_dict[r]:
                # skip if no tid in this rank
                if ttid=="0":
                    continue
                return ttid

    return '1'

def acc2taxid_raw(acc: str, accession2taxid_file: Optional[str] = None) -> str:
    """
    Get the taxonomy ID for a given accession from NCBI accession2taxid tsv file.

    Args:
        acc (str): The accession number to look up.
        accession2taxid_file (str, optional): The path to the accession to taxonomy ID mapping file. If not specified, the default file in the taxonomy directory will be used.

    Returns:
        str: The taxonomy ID for the given accession.
    """

    # Remove version number
    acc = acc.split('.')[0]

    if not acc in accTid:
        logger.info( f"acc2taxid from file: {accession2taxid_file}" )
        with open( accession2taxid_file ) as f:
            f.seek(0, 2)
            start = 0
            end = f.tell()
            accCur = ""
            
            while( acc != accCur and start < end ):
                posNew = (end+start)/2
                f.seek( posNew )
                if posNew != start: f.readline()
                line = f.readline()    
                
                logger.debug( "start: %15d, posNew: %15d, end: %15d, line: %s" % (start, posNew, end, line) )
                if line :
                    (accNew, accNewVer, tid, gi) = line.split('\t')
                else:
                    break

                logger.debug( f'[acc, accNew, accCur]=[{acc}, {accNew}, {accCur}]')
                if acc > accNew and accCur != accNew and accNew:
                    if accNew: posNew = f.tell()
                    start = posNew
                    if start >= end: end = start+1
                else:
                    end = posNew
                
                accCur = accNew
            f.close()

            if accCur == acc:
                accTid[acc] = tid.strip()
            else:
                accTid[acc] = ""

    return accTid[acc]

def acc2taxid(acc: str, type: Optional[str]='nucl') -> str:
    """
    Get the taxonomy ID for a given accession.

    Args:
        acc (str): The accession number to look up.
        type (str, optional): Type of the acession number, either nucl, prot, or pdb. Default is 'nucl'.

    Returns:
        str: The taxonomy ID for the given accession.
    """
    global taxonomy_dir
    import glob
    acc2taxid_files = []
    
    # preparing accession2taxid files
    if type == 'nucl':
        acc2taxid_files = [
            f'{taxonomy_dir}/accession2taxid/nucl_gb.accession2taxid',
            f'{taxonomy_dir}/accession2taxid/nucl_wgs.accession2taxid.EXTRA',
            f'{taxonomy_dir}/accession2taxid/nucl_wgs.accession2taxid',
            f'{taxonomy_dir}/accession2taxid/dead_nucl.accession2taxid',
            f'{taxonomy_dir}/accession2taxid/dead_wgs.accession2taxid',
        ]
    elif type == 'prot':
        acc2taxid_files = [
            f'{taxonomy_dir}/accession2taxid/prot.accession2taxid.FULL',
            f'{taxonomy_dir}/accession2taxid/dead_prot.accession2taxid',
        ]
    elif type == 'pdb':
        acc2taxid_files = [
            f'{taxonomy_dir}/accession2taxid/pdb.accession2taxid'
        ]

    # check if accession2taxid files exist
    avail_acc2taxid_files = []

    for acc2taxid_file in acc2taxid_files:
        logger.debug( f"checking {acc2taxid_file}" )
        if os.path.isfile(acc2taxid_file):
            avail_acc2taxid_files.append(acc2taxid_file)
    
    logger.debug( f"avail_acc2taxid_files: {avail_acc2taxid_files}" )

    # download accession2taxid files if not exist
    if len(avail_acc2taxid_files) == 0:
        logger.info( f"NCBI accession2taxid data not found. Please run `detaxa update --help` for details." )
        print( f"WARNING: NCBI accession2taxid data not found. Please run `detaxa update --help` for details." )

    for acc2taxid_file in avail_acc2taxid_files:
        taxid = acc2taxid_raw(acc, accession2taxid_file=acc2taxid_file)
        if taxid: return taxid

    return ""

def taxid2decendentOnRank(tid: Union[int, str], target_rank=None) -> list:
    """
    Return a list of taxids for all descendants of the given taxid at the specified target rank.

    Arguments:
    - tid: The taxid to retrieve the descendants for. Can be either an integer or a string.
    - target_rank: The rank at which to retrieve the descendants. Defaults to None, which means all descendants will be returned.

    Returns:
    - A list of taxids for all descendants of the given taxid at the specified target rank.
    """
    tid = _checkTaxonomy(tid)
    decd_tids = [k for k, v in taxParents.items() if v == tid]

    tids = []

    if not target_rank:
        return decd_tids
    else:
        for tid in decd_tids:
            if target_rank == _getTaxRank(tid):
                tids.append(tid)
            else:
                if taxidIsLeaf(tid):
                    continue
                else:
                    tids.extend(taxid2decendentOnRank(tid, target_rank))

        return tids

def loadTaxonomy(dbpath: Optional[str] = None,
                 cus_taxonomy_file: Optional[str] = None, 
                 cus_taxonomy_format: str = 'tsv',
                 auto_download: bool = True) -> None:
    """
    Load taxonomy files into memory for use in subsequent conversions.

    Args:
        dbpath (str, optional): Path to store and load the taxonomy files. Defaults to None.
        cus_taxonomy_file (str, optional): Path to a custom taxonomy file. Defaults to None.
        cus_taxonomy_format (str, optional): Format of the custom taxonomy file, one of ['tsv','mgnify_lineage','gtdb_taxonomy','gtdb_metadata']. Defaults to 'tsv'.
        auto_download (bool, optional): If True, automatically download the taxonomy files if they are not found locally. Defaults to True.

    Returns:
        None
    """
    global taxonomy_dir, abbr_json_path
    
    logger.debug( f"v{__version__}" )

    if dbpath:
        if os.path.isdir(dbpath):
            taxonomy_dir = dbpath
            if os.path.isfile( f"{taxonomy_dir}/major_level_to_abbr.json" ):
                abbr_json_path = f"{taxonomy_dir}/major_level_to_abbr.json"
        else:
            logger.info( f"Invalid dbpath: {dbpath} is not a directory. Default dbpath will be used." )

    logger.debug( f"Taxonomy directory: {taxonomy_dir}" )

    # loading major levels to json file
    _loadAbbrJson(abbr_json_path)

    #NCBI ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/taxdump.tar.gz
    taxdump_tgz_file = taxonomy_dir+"/taxdump.tar.gz"

    #raw taxonomy dmp files from NCBI
    names_dmp_file = taxonomy_dir+"/names.dmp"
    nodes_dmp_file = taxonomy_dir+"/nodes.dmp"
    merged_dmp_file = taxonomy_dir+"/merged.dmp"

    #parsed taxonomy tsv file
    taxonomy_file = taxonomy_dir+"/taxonomy.tsv"
    merged_taxonomy_file = taxonomy_dir+"/taxonomy.merged.tsv"

    #custom taxonomy file
    if not cus_taxonomy_file:
        cus_taxonomy_file = taxonomy_dir+"/taxonomy.custom.tsv"

    # checking if taxonomy files provided
    if not os.path.isfile( taxdump_tgz_file ) \
        and not os.path.isfile( merged_taxonomy_file ) \
        and not os.path.isfile( taxonomy_file ) \
        and not (os.path.isfile( names_dmp_file ) and os.path.isfile( nodes_dmp_file )) \
        and not os.path.isfile( cus_taxonomy_file ):
        
        logger.info( f"No taxonomy files not found." )
        if auto_download:
            NCBITaxonomyDownload(taxonomy_dir)
        else:
            logger.info( f"Auto-download is off." )
            logger.fatal( f"No available taxonomy files." )
            _die( "[ERROR] No available taxonomy files." )

    # try to load taxonomy from taxonomy.tsv
    if os.path.isfile( nodes_dmp_file ) and os.path.isfile( names_dmp_file ):
        loadNCBITaxonomy(taxdump_tgz_file, names_dmp_file, nodes_dmp_file, merged_dmp_file)
    elif os.path.isfile(taxdump_tgz_file):
        loadNCBITaxonomy(taxdump_tgz_file, names_dmp_file, nodes_dmp_file, merged_dmp_file)

    if os.path.isfile(taxonomy_file):
        logger.info( "Open taxonomy file: %s"% taxonomy_file )
        loadTaxonomyTSV(taxonomy_file)

    # try to load custom taxonomy from taxonomy.custom.tsv
    if os.path.isfile(cus_taxonomy_file) and (cus_taxonomy_format=='tsv'):
        logger.info( "Open custom taxonomy node file (tsv format): %s"% cus_taxonomy_file)
        loadTaxonomyTSV(cus_taxonomy_file)
    # try to load custom taxonomy from lineage file
    elif os.path.isfile(cus_taxonomy_file) and (cus_taxonomy_format=='mgnify_lineage'):
        logger.info( "Open custom taxonomy node file (lineage format): %s"% cus_taxonomy_file)
        loadMgnifyTaxonomy(cus_taxonomy_file)
    # try to load custom taxonomy from GTDB file
    elif os.path.isfile(cus_taxonomy_file) and (cus_taxonomy_format in ['gtdb_taxonomy','gtdb_metadata']):
        loadGTDBTaxonomy(cus_taxonomy_file, cus_taxonomy_format)
    elif os.path.isfile(cus_taxonomy_file):
        logger.fatal( f"invalid cus_taxonomy_format: {cus_taxonomy_format}" )
        _die(f"[ERROR] Invalid cus_taxonomy_format: {cus_taxonomy_format}")

def NCBITaxonomyDownload(dir=None, taxdump=True, acc_wgs=False, acc_nucl=False, acc_prot=False, acc_pdb=False, acc_dead=True):
    import requests
    global taxonomy_dir

    if not dir:
        dir = taxonomy_dir
        logger.info( f"No destination taxonomy dir input. Use default: {dir}..." )

    if not os.path.exists( dir ):
        os.makedirs( dir )
        logger.info( f"Taxonomy dir doesn't exist. Make dir: {dir}..." )

    if taxdump:
        url = 'http://ftp.ncbi.nlm.nih.gov/pub/taxonomy/taxdump.tar.gz'
        logger.info( f"Auto downloading taxanomy from {url}..." )
        # download taxonomy file if auto_download enabled
        r = requests.get(url)
        
        taxdump_tgz_file = f'{dir}/taxdump.tar.gz'

        with open(taxdump_tgz_file, 'wb') as f:
            f.write(r.content)
        if os.path.getsize( taxdump_tgz_file ):
            logger.info( f"Saved to {taxdump_tgz_file}." )
        else:
            logger.fatal( f"Failed to download or save taxonomy files." )
            _die( "[ERROR] Failed to download or save taxonomy files." )    

        # extract
        tax_tar = tarfile.open(taxdump_tgz_file, "r:gz")
        logger.info( f"Extracting nodes.dmp..." )
        tax_tar.extract('nodes.dmp', dir)
        logger.info( f"Extracting names.dmp..." )
        tax_tar.extract('names.dmp', dir)
        logger.info( f"Extracting merged.dmp..." )
        tax_tar.extract('merged.dmp', dir)
        tax_tar.close()
        # # delete taxdump_tgz_file
        # os.remove(taxdump_tgz_file)

    if acc_wgs or acc_nucl or acc_prot or acc_pdb:
        import subprocess
        url = "rsync://ftp.ncbi.nlm.nih.gov/pub/taxonomy/accession2taxid"
        
        args = ["rsync", "-avh"]
        if not acc_dead:
            args.append("--exclude='*dead*'")
        if not acc_wgs:
            args.append("--exclude='*wgs.*'")
        if not acc_prot:
            args.append("--exclude='*prot.*'")
        if not acc_nucl:
            args.append("--exclude='*nucl.*'")
        if not acc_pdb:
            args.append("--exclude='*pdb.*'")

        cmd = " ".join(args+[url, dir])
        logger.debug(cmd)
        logger.info( f"Auto ryncing accession2taxid data from {url}..." )
        subprocess.call(cmd, shell=True)

        logger.info( f"Decompressing accession2taxid data..." )
        cmd = f"gzip -f -d {dir}/accession2taxid/*.gz"
        subprocess.call(cmd, shell=True)
    
    logger.info( f"Done." )

def loadTaxonomyTSV(tsv_taxonomy_file):

    # loading major levels to json file
    _loadAbbrJson(abbr_json_path)

    try:
        with open(tsv_taxonomy_file) as f:
            for line in f:
                line = line.rstrip('\r\n')
                if not line: continue
                tid, depth, parent, rank, name = line.split('\t')
                taxParents[tid] = parent
                taxDepths[tid] = int(depth)
                taxRanks[tid] = rank
                taxNames[tid] = name
                if parent in taxNumChilds:
                    taxNumChilds[parent] += 1
                else:
                    taxNumChilds[parent] = 1
            f.close()
            logger.info( f"Done parsing custom tsv taxonomy file." )
    except IOError:
        _die( "Failed to open custom tsv taxonomy file: %s." % tsv_taxonomy_file )

def loadNCBITaxonomy(taxdump_tgz_file: Optional[str] = None, 
                     names_dmp_file: Optional[str] = None, 
                     nodes_dmp_file: Optional[str] = None, 
                     merged_dmp_file: Optional[str] = None):

    # loading major levels from json file
    _loadAbbrJson(abbr_json_path)

    # try to load taxonomy from taxonomy.tsv
    if os.path.isfile(nodes_dmp_file) and os.path.isfile(names_dmp_file):
        try:
            # read name from names.dmp
            logger.info( f"Open taxonomy name file: {names_dmp_file}" )
            with open(names_dmp_file) as f:
                for line in f:
                    tid, name, tmp, nametype = line.rstrip('\r\n').split('\t|\t')
                    if not nametype.startswith("scientific name"):
                        continue
                    taxNames[tid] = name
                f.close()
                logger.info( f"Done parsing taxonomy name file." )    

            # read taxonomy info from nodes.dmp
            logger.info( f"Open taxonomy node file: {nodes_dmp_file}" )
            with open(nodes_dmp_file) as f:
                for line in f:
                    fields = line.rstrip('\r\n').split('\t|\t')
                    tid = fields[0]
                    parent = fields[1]
                    taxParents[tid] = parent
                    taxDepths[tid] = int(taxDepths[parent])+1 if parent in taxDepths else 0 # could have potiential bug if child node is parsed before parent node.
                    taxRanks[tid] = fields[2]
                    if parent in taxNumChilds:
                        taxNumChilds[parent] += 1
                    else:
                        taxNumChilds[parent] = 1
                f.close()
                logger.info( f"Done parsing taxonomy node file." )
        except IOError:
            _die( "Failed to open taxonomy files (taxonomy.tsv, nodes.dmp and names.dmp)." )
    elif os.path.isfile( taxdump_tgz_file ):
        try:
            logger.info( f"Open taxonomy file: {taxdump_tgz_file}" )
            tar = tarfile.open(taxdump_tgz_file, "r:gz")
            
            # read name from names.dmp
            logger.info( "Extract taxonomy names file: names.dmp" )
            member = tar.getmember("names.dmp")
            f = tar.extractfile(member)
            for line in f.readlines():
                tid, name, tmp, nametype = line.decode('utf8').rstrip('\r\n').split('\t|\t')
                if not nametype.startswith("scientific name"):
                    continue
                taxNames[tid] = name
            f.close()
            
            # read taxonomy info from nodes.dmp
            logger.info( "Extract taxonomy nodes file: nodes.dmp" )
            member = tar.getmember("nodes.dmp")
            f = tar.extractfile(member)
            for line in f.readlines():
                fields = line.decode('utf8').rstrip('\r\n').split('\t|\t')
                tid = fields[0]
                parent = fields[1]
                taxParents[tid] = parent
                taxDepths[tid] = int(taxDepths[parent])+1 if parent in taxDepths else 0 # could have potiential bug if child node is parsed before parent node.
                taxRanks[tid] = fields[2]
                if parent in taxNumChilds:
                    taxNumChilds[parent] += 1
                else:
                    taxNumChilds[parent] = 1
            f.close()

            # read taxonomy info from merged.dmp
            logger.info( "Extract taxonomy merged file: merged.dmp" )
            member = tar.getmember("merged.dmp")
            f = tar.extractfile(member)
            for line in f.readlines():
                fields = line.decode('utf8').rstrip('\r\n').split('\t|')
                taxMerged[fields[0]] = fields[1].strip('\t')
            f.close()

        except IOError:
            _die( "Failed to load taxonomy from %s"%taxdump_tgz_file )
    
    #try to load merged taxids
    if os.path.isfile( merged_dmp_file ):
        logger.info( "Open merged taxonomy node file: %s"% merged_dmp_file )
        with open(merged_dmp_file) as f:
            for line in f:
                fields = line.rstrip('\r\n').split('\t|')
                taxMerged[fields[0]] = fields[1].strip('\t')
            f.close()
            logger.info( f"Done parsing merged taxonomy file." )
    
def loadMgnifyTaxonomy(mgnify_taxonomy_file=None):
    """
    loadMgnifyTaxonomy()
    """

    # loading major levels to json file
    _loadAbbrJson(abbr_json_path)

    # try to load custom taxonomy from lineage file
    if os.path.isfile(mgnify_taxonomy_file):
        logger.info( "Open custom taxonomy node file (lineage format): %s"% mgnify_taxonomy_file)
        try:
            with open(mgnify_taxonomy_file) as f:
                for line in f:
                    line = line.rstrip('\r\n')
                    if not line: continue
                    if line.startswith('#'): continue
                    if not line.startswith('sk__'):
                        logger.warn( f"A text line of lineage has to start with 'sk__'...skipped: {line}" )
                        continue

                    temp = line.split(';')
                    p_name = ''
                    rank = ''
                    name = ''

                    for i in range(1, len(temp)+1):
                        # this taxa
                        (rank_abbr, name) = temp[-i].split('__')

                        import re
                        re_taxa = re.compile("^([^_]+)__(.*)$")
                        rank_abbr, name = re_taxa.match(temp[-i]).groups()

                        # for na taxon (no_{rank_abbr}_rank)
                        if name=="": name = p_name

                        # paranet taxa
                        try:
                            p_rank_abbr, p_name = re_taxa.match(temp[-(i+1)]).groups()
                            if p_name=="":
                                p_name = f'{name} - no_{rank_abbr}_rank'
                        except:
                            # for the superkingdom rank, assign parant taxid to 1 (root)
                            p_name = '1'
                            if not '1' in taxRanks: taxRanks['1'] = 'root'
                            if not '1' in taxNames: taxNames['1'] = 'root'

                        if rank_abbr in abbr_to_major_level:
                            rank = abbr_to_major_level[rank_abbr]
                        else:
                            rank = rank_abbr
                            
                        tid = name
                        taxParents[tid] = p_name
                        taxRanks[tid] = rank
                        taxNames[tid] = name
                        if p_name in taxNumChilds:
                            taxNumChilds[p_name] += 1
                        else:
                            taxNumChilds[p_name] = 1
                f.close()
                logger.info( f"Done parsing custom taxonomy file." )
        except IOError:
            _die( "Failed to open custom taxonomy file: %s." % mgnify_taxonomy_file )

    logger.info( f"Done parsing taxonomy files (total {len(taxParents)} taxa loaded)" )

def loadGTDBTaxonomy(gtdb_taxonomy_file=None, gtdb_taxonomy_format="gtdb_metadata"):
    """
    loadGTDBTaxonomy()
    """

    # loading major levels to json file
    _loadAbbrJson(abbr_json_path, type='gtdb')

    # try to load custom taxonomy from GTDB file
    if os.path.isfile(gtdb_taxonomy_file) and (gtdb_taxonomy_format in ['gtdb_taxonomy','gtdb_metadata']):
        logger.info( f"Open custom taxonomy node file ({gtdb_taxonomy_format}): %s"% gtdb_taxonomy_file)
        try:
            with open(gtdb_taxonomy_file) as f:
                for line in f:
                    line = line.rstrip('\r\n')
                    if not line: continue
                    if line.startswith('#'): continue
                    if line.startswith('accession'): continue

                    acc, lineage = '', ''

                    if gtdb_taxonomy_format=='gtdb_taxonomy':
                        try:
                            acc, lineage = line.split('\t')
                        except:
                            logger.fatal( f"Incorrect GTDB taxonomy .tsv format: {gtdb_taxonomy_file}" )
                            _die( f"[ERROR] 2 columns are required for GTDB taxonomy .tsv format: {gtdb_taxonomy_file}" )
                        lineage = f'{lineage};x__{acc}'
                    elif gtdb_taxonomy_format=='gtdb_metadata':
                        temp = line.split('\t')
                        if len(temp)!=110:
                            logger.fatal( f"Incorrect GTDB metadata .tsv format: {gtdb_taxonomy_file}" )
                            _die( f"[ERROR] 110 columns are required for GTDB metadata .tsv format: {gtdb_taxonomy_file}" )
                        acc = temp[0]
                        # col_17: 'gtdb_taxonomy'; col_63: 'ncbi_organism_name'
                        lineage = f'{temp[16]};x__{temp[62]}'
                    else:
                        logger.fatal( f"Incorrect format: {gtdb_taxonomy_format}: {gtdb_taxonomy_file}" )
                        _die( f"[ERROR] Incorrect format: {gtdb_taxonomy_format}: {gtdb_taxonomy_file}" )

                    temp = lineage.split(';')
                    p_name = ''
                    rank = ''
                    name = ''

                    for i in range(1, len(temp)+1):
                        # this taxa
                        import re
                        re_taxa = re.compile("^([^_]+)__(.*)$")
                        rank_abbr, name = re_taxa.match(temp[-i]).groups()
                        if i==1 and rank_abbr=='x':
                            name = f'{name} ({acc})'
                        # for na taxon (no_{rank_abbr}_rank)
                        if name=="": name = p_name

                        # paranet taxa
                        try:
                            p_rank_abbr, p_name = re_taxa.match(temp[-(i+1)]).groups()
                            if p_name=="":
                                p_name = f'{name} - no_{p_rank_abbr}_rank'
                        except:
                            # for the *first* taxa in lineage line (usually superkingdom), assign parant taxid to 1 (root)
                            p_name = '1'
                            if not '1' in taxRanks: taxRanks['1'] = 'root'
                            if not '1' in taxNames: taxNames['1'] = 'root'

                        if rank_abbr=='d':
                            rank = 'superkingdom'
                        elif rank_abbr=='x':
                            rank = 'strain'
                        if rank_abbr in abbr_to_major_level:
                            rank = abbr_to_major_level[rank_abbr]
                        else:
                            rank = rank_abbr
                            
                        tid = acc.split('.')[0] if i==1 and rank=='strain' else name
                        taxParents[tid] = p_name
                        taxRanks[tid] = rank
                        taxNames[tid] = name
                        if p_name in taxNumChilds:
                            taxNumChilds[p_name] += 1
                        else:
                            taxNumChilds[p_name] = 1
                f.close()
                logger.info( f"Done parsing custom taxonomy file." )
        except IOError:
            _die( "Failed to open custom taxonomy file: %s." % gtdb_taxonomy_file )

    logger.info( f"Done parsing taxonomy files (total {len(taxParents)} taxa loaded)" )




#!/usr/bin/env python3
import sys
import io
from gottcha.scripts import pull_database
from gottcha.scripts import gottcha2

def usage():
    print("usage: gottcha2 profile --help\ngottcha2 pull --help")
    sys.exit()

def gottcha2_command():
    args = sys.argv[1:]
    if len(args) < 1:
        usage()
    elif args[0] == "profile":
        gottcha2.main(args[1:])
    elif args[0] == "pull":
        pull_database.main(args[1:])
    else:
        print("%s is not a valid argument", args[1])
        usage()
